/*
 *  Copyright (C) 2003 Patrizio Bruno.
 *  E-mail: desertconsulting@firenze.net.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111 USA
 */

#include <string.h>
#include <gtk/gtk.h>

#include "gtkmdi.h"
#include "gtkmdi_internal.h"

extern guint gtk_mdi_signals[LAST_SIGNAL];

/* private */
static void gtk_mdi_window_class_init (GtkMdiWindowClass *);
static void gtk_mdi_window_init (GtkMdiWindow *);
static void gtk_mdi_window_finalize (GObject *);
static void gtk_mdi_window_destroy (GtkObject *);
static void gtk_mdi_window_size_request (GtkWidget *widget,
					GtkRequisition *requisition);
static void gtk_mdi_window_size_allocate (GtkWidget *widget,
					GtkAllocation *allocation);
static gboolean gtk_mdi_window_expose (GtkWidget *, GdkEventExpose *);
static void gtk_mdi_window_realize (GtkWidget *);
static void gtk_mdi_window_unrealize (GtkWidget *);
static gboolean gtk_mdi_window_real_focus (GtkWidget *, GtkDirectionType);
static void gtk_mdi_window_state_changed (GtkWidget *, GtkStateType);
static void gtk_mdi_window_style_set (GtkWidget *, GtkStyle *);
static void gtk_mdi_window_add (GtkContainer *, GtkWidget *);
static void gtk_mdi_window_remove (GtkContainer *, GtkWidget *);

/* inrternals */
static void _gtk_mdi_window_show_borders (GtkMdiWindow *window);
static void _gtk_mdi_window_hide_borders (GtkMdiWindow *window);
static void _gtk_mdi_window_show_titlebar (GtkMdiWindow *window);
static void _gtk_mdi_window_hide_titlebar (GtkMdiWindow *window);
static void _gtk_mdi_window_load_borders (GtkEventBox *tlc, GtkEventBox *top,
			GtkEventBox *trc, GtkEventBox *left,
			GtkEventBox *right, GtkEventBox *blc,
			GtkEventBox *bottom, GtkEventBox *brc);
static void _gtk_mdi_window_resize_border (gpointer data, gpointer user_data);
static void _gtk_mdi_window_resize_borders (GtkMdiWindow *window);
static gint _gtk_mdi_window_has_focus_compare (gpointer a, gpointer b);
gboolean _gtk_mdi_window_maximize (GtkWidget *widget, GtkWidget *data);
gboolean _gtk_mdi_window_close (GtkWidget *widget, GtkWidget *data);
gboolean _gtk_mdi_window_iconify (GtkWidget *widget, GtkWidget *data);

static GtkEventBoxClass *parent_class;

struct _GtkMdiWindowPrivate
{
	GtkLabel *title_label;
	GtkHBox *title_bar;
	GtkImage *icon;
	GtkWidget *icon_holder;
	GtkMdiWindowState state;
	GtkMdiWindowState old_state;
	GtkEventBox *container;
	GSList *borders;
	guint old_width;
	guint old_height;
	guint old_pos_x;
	guint old_pos_y;
	gboolean has_focus;
	gboolean accept_focus;
	gboolean resizeable;
	gboolean skip_taskbar;
	gboolean skip_pager;
	gboolean decorated;
};

GType
gtk_mdi_window_get_type (void)
{
	static GType gtk_mdi_window_type = 0;

	if (!gtk_mdi_window_type)
	{
		static const GTypeInfo gtk_mdi_window_info = {
			sizeof (GtkMdiWindowClass),
			NULL,
			NULL,
			(GClassInitFunc) gtk_mdi_window_class_init,
			NULL,
			NULL,
			sizeof (GtkMdiWindow),
			0,
			(GInstanceInitFunc) gtk_mdi_window_init
		};

		gtk_mdi_window_type = g_type_register_static (
							GTK_TYPE_EVENT_BOX,
							"GtkMdiWindow",
							&gtk_mdi_window_info,
							0);
	}

	return gtk_mdi_window_type;
}

static void
gtk_mdi_window_class_init (GtkMdiWindowClass *class)
{
	GObjectClass *o_class;
	GtkObjectClass *gtko_class;
	GtkWidgetClass *w_class;
	GtkContainerClass *c_class;

	parent_class = g_type_class_peek_parent (class);
	o_class = (GObjectClass *) class;
	gtko_class = (GtkObjectClass *) class;
	w_class = (GtkWidgetClass *) class;
	c_class = (GtkContainerClass *) class;

	o_class->finalize = gtk_mdi_window_finalize;
	gtko_class->destroy = gtk_mdi_window_destroy;

	w_class->size_request = gtk_mdi_window_size_request;
	w_class->size_allocate = gtk_mdi_window_size_allocate;
	w_class->realize = gtk_mdi_window_realize;
	w_class->unrealize = gtk_mdi_window_unrealize;
	//w_class->focus = gtk_mdi_window_real_focus;
	//w_class->state_changed = gtk_mdi_window_state_changed;
	//w_class->style_set = gtk_mdi_window_style_set;
	w_class->expose_event = gtk_mdi_window_expose;

	c_class->add = gtk_mdi_window_add;
	c_class->remove = gtk_mdi_window_remove;
}

static void
gtk_mdi_window_destroy (GtkObject *object)
{
	g_return_if_fail (GTK_IS_MDI_WINDOW (object));

	if (GTK_OBJECT_CLASS (parent_class)->destroy)
		GTK_OBJECT_CLASS (parent_class)->destroy (object);
}

static void
gtk_mdi_window_size_request (GtkWidget *widget, GtkRequisition *requisition)
{
	g_return_if_fail (GTK_IS_MDI_WINDOW (widget));
	g_return_if_fail (requisition != NULL);

	if (GTK_WIDGET_CLASS (parent_class)->size_request)
		GTK_WIDGET_CLASS (parent_class)->size_request (widget,
								requisition);

	_gtk_mdi_window_resize_borders (GTK_MDI_WINDOW (widget));
}

static void
gtk_mdi_window_size_allocate (GtkWidget *widget, GtkAllocation *allocation)
{
	g_return_if_fail (GTK_IS_MDI_WINDOW (widget));
	g_return_if_fail (allocation != NULL);

	GTK_WIDGET_CLASS (parent_class)->size_allocate(widget,
			allocation);
}

static void
_gtk_mdi_window_resize_borders (GtkMdiWindow *window)
{
	g_return_if_fail (GTK_IS_MDI_WINDOW (window));

	g_slist_foreach (window->private->borders,
			_gtk_mdi_window_resize_border,
			window);
}

static void
_gtk_mdi_window_resize_border (gpointer data, gpointer user_data)
{
	GtkWidget *widget = GTK_WIDGET (user_data);
	GtkWidget *ebox = GTK_WIDGET (data);
	GtkImage *icon;
	GdkPixbuf *pbuf = NULL;
	GdkImage *image = NULL;
	GdkBitmap *mask = NULL;
	gint w = -1;
	gint h = -1;

	icon = GTK_IMAGE (gtk_bin_get_child (GTK_BIN (ebox)));

	g_return_if_fail (GTK_IS_IMAGE (icon));

	switch (gtk_image_get_storage_type (icon))
	{
		case GTK_IMAGE_EMPTY:
		case GTK_IMAGE_STOCK:
		case GTK_IMAGE_ICON_SET:
		case GTK_IMAGE_ANIMATION:
		case GTK_IMAGE_PIXMAP:
			break;
		case GTK_IMAGE_PIXBUF:
			pbuf = gtk_image_get_pixbuf (icon);
			break;
		case GTK_IMAGE_IMAGE:
			gtk_image_get_image (icon, &image, &mask);
			break;
	}

	if(image)
	{
		pbuf = gdk_pixbuf_get_from_image (NULL, image, NULL,
						0, 0,
						0, 0,
						image->width, image->height);
		w = image->width;
		h = image->height;
	}
	else if(pbuf)
	{
		w = gdk_pixbuf_get_width (pbuf);
		h = gdk_pixbuf_get_height (pbuf);
	}
	else
	{
		pbuf = gdk_pixbuf_get_from_drawable (NULL,
						GTK_WIDGET (icon)->window, NULL,
						0, 0,
						0, 0,
						image->width, image->height);
		w = 0;
		h = 0;
		gdk_drawable_get_size (GTK_WIDGET (icon)->window, &w, &h);
		if (!w && !h)
		{
			w = 0;
			h = 0;
		}
	}

	if (pbuf && (w || h))
	{
		GdkPixbuf *npb;
		gint dest_w = 0;
		gint dest_h = 0;
		GtkWidget *new_image;

		if ((w == 4 ) && (h != 4))
		{
			dest_w = w;
			dest_h = widget->requisition.height - 10;
			if (dest_h < 0) dest_h = 80;
		}
		else if((w != 4) && (h == 4))
		{
			dest_w = widget->requisition.width - 10;
			if (dest_w < 0) dest_w = 80;
			dest_h = h;
		}
		else return;

		npb = gdk_pixbuf_scale_simple (pbuf,
						dest_w, dest_h,
						GDK_INTERP_HYPER);
		new_image = gtk_image_new_from_pixbuf (npb);
		gtk_container_remove (GTK_CONTAINER (ebox), GTK_WIDGET (icon));
		gtk_widget_show (new_image);
		gtk_container_add (GTK_CONTAINER (ebox), new_image);
	}
}

static gboolean
gtk_mdi_window_expose (GtkWidget *widget, GdkEventExpose *event)
{
	if (GTK_WIDGET_CLASS (parent_class)->expose_event)
		(GTK_WIDGET_CLASS (parent_class)->expose_event) (widget,
								event);

	return TRUE;
}

static void
gtk_mdi_window_realize (GtkWidget *widget)
{
	g_return_if_fail (GTK_IS_MDI_WINDOW (widget));

	if (GTK_WIDGET_CLASS (parent_class)->realize)
		GTK_WIDGET_CLASS (parent_class)->realize (widget);
}

static void
gtk_mdi_window_unrealize (GtkWidget *widget)
{
	g_return_if_fail (GTK_IS_MDI_WINDOW (widget));

	if (GTK_WIDGET_CLASS (parent_class)->unrealize)
		GTK_WIDGET_CLASS (parent_class)->unrealize (widget);
}

static gboolean
gtk_mdi_window_real_focus (GtkWidget *widget, GtkDirectionType direction)
{
	g_return_val_if_fail (GTK_IS_MDI_WINDOW (widget), FALSE);

	if (GTK_WIDGET_CAN_FOCUS (widget))
	{
		GTK_WIDGET_UNSET_FLAGS (widget, GTK_CAN_FOCUS);
	}
	return GTK_WIDGET_CLASS (parent_class)->focus (widget, direction);
}

static void
gtk_mdi_window_state_changed (GtkWidget *widget, GtkStateType previous_state)
{
}

static void
gtk_mdi_window_style_set (GtkWidget *widget, GtkStyle *previous_style)
{
}

static void
gtk_mdi_window_add (GtkContainer *container, GtkWidget *widget)
{
	g_return_if_fail (GTK_IS_MDI_WINDOW (container));
	g_return_if_fail (GTK_IS_WIDGET (widget));

	if (GTK_IS_MDI (GTK_MDI_WINDOW (container)->mdicontainer))
	{
		//gtk_widget_show (widget);
		//gtk_widget_modify_bg (widget, GTK_STATE_NORMAL, &(GTK_WIDGET (GTK_MDI_WINDOW (container)->mdicontainer)->parent->style->base [GTK_STATE_ACTIVE]));
	}

	if (GTK_IS_EVENT_BOX (GTK_MDI_WINDOW (container)->private->container))
		gtk_container_add (
		GTK_CONTAINER (GTK_MDI_WINDOW (container)->private->container),
			widget);
	else  if (GTK_CONTAINER_CLASS (parent_class)->remove)
		GTK_CONTAINER_CLASS (parent_class)->add(container, widget);
}

static void
gtk_mdi_window_remove (GtkContainer *container, GtkWidget *widget)
{
	g_return_if_fail (GTK_IS_MDI_WINDOW (container));
	g_return_if_fail (GTK_IS_WIDGET (widget));

	if (GTK_IS_EVENT_BOX (GTK_MDI_WINDOW (container)->private->container))
		gtk_container_remove (
		GTK_CONTAINER (GTK_MDI_WINDOW (container)->private->container),
			widget);
	else if (GTK_CONTAINER_CLASS (parent_class)->remove)
		GTK_CONTAINER_CLASS (parent_class)->remove (container,
								widget);
}

static void
gtk_mdi_window_init (GtkMdiWindow *window)
{
	GtkTable *table;
	GtkEventBox *top_left_corner;
	GtkEventBox *top_right_corner;
	GtkEventBox *bottom_left_corner;
	GtkEventBox *bottom_right_corner;
	GtkEventBox *top;
	GtkEventBox *right;
	GtkEventBox *left;
	GtkEventBox *bottom;
	GtkVBox *center;
	GtkWidget *close_butt;
	GtkWidget *max_butt;
	GtkWidget *icon_butt;

	window->private = g_new0 (GtkMdiWindowPrivate, 1);

	close_butt = gtk_button_new();
	gtk_container_add (GTK_CONTAINER (close_butt), 
		gtk_image_new_from_stock ("gtk-close", GTK_ICON_SIZE_MENU));
	g_signal_connect (close_butt, "released",
			_gtk_mdi_window_close, window);

	max_butt = gtk_button_new();
	gtk_container_add (GTK_CONTAINER (max_butt), 
		gtk_image_new_from_stock ("gtk-zoom-fit", GTK_ICON_SIZE_MENU));
	g_signal_connect (max_butt, "released",
			_gtk_mdi_window_maximize, window);

	icon_butt = gtk_button_new();
	gtk_container_add (GTK_CONTAINER (icon_butt), 
		gtk_image_new_from_stock ("gtk-goto-bottom",
					GTK_ICON_SIZE_MENU));
	g_signal_connect (icon_butt, "released",
			_gtk_mdi_window_iconify, window);

	top_left_corner = GTK_EVENT_BOX (gtk_event_box_new ());
	top_right_corner = GTK_EVENT_BOX (gtk_event_box_new ());
	bottom_left_corner = GTK_EVENT_BOX (gtk_event_box_new ());
	bottom_right_corner = GTK_EVENT_BOX (gtk_event_box_new ());
	top = GTK_EVENT_BOX (gtk_event_box_new ());
	bottom = GTK_EVENT_BOX (gtk_event_box_new ());
	left = GTK_EVENT_BOX (gtk_event_box_new ());
	right = GTK_EVENT_BOX (gtk_event_box_new ());

	_gtk_mdi_window_load_borders (
				top_left_corner,
				top,
				top_right_corner,
				left,
				right,
				bottom_left_corner,
				bottom,
				bottom_right_corner
			);

	gtk_widget_show (GTK_WIDGET (top_left_corner));
	gtk_widget_show (GTK_WIDGET (top_right_corner));
	gtk_widget_show (GTK_WIDGET (bottom_left_corner));
	gtk_widget_show (GTK_WIDGET (bottom_right_corner));
	gtk_widget_show (GTK_WIDGET (top));
	gtk_widget_show (GTK_WIDGET (bottom));
	gtk_widget_show (GTK_WIDGET (left));
	gtk_widget_show (GTK_WIDGET (right));

	center = GTK_VBOX (gtk_vbox_new (FALSE, 2));
	gtk_widget_show (GTK_WIDGET (center));
	table = GTK_TABLE (gtk_table_new(3, 3, FALSE));
	gtk_widget_show (GTK_WIDGET (table));

	gtk_table_attach (table, GTK_WIDGET (top_left_corner),
			0, 1, 0, 1,
			(GtkAttachOptions)0, (GtkAttachOptions)0,
			0, 0);
	gtk_table_attach (table, GTK_WIDGET (top),
			1, 2, 0, 1,
			GTK_FILL | GTK_EXPAND | GTK_SHRINK,
			(GtkAttachOptions)0,
			0, 0);
	gtk_table_attach (table, GTK_WIDGET (top_right_corner),
			2, 3, 0, 1,
			(GtkAttachOptions)0, (GtkAttachOptions)0,
			0, 0);
	gtk_table_attach (table, GTK_WIDGET (left),
			0, 1, 1, 2,
			(GtkAttachOptions)0,
			GTK_FILL | GTK_EXPAND | GTK_SHRINK,
			0, 0);
	gtk_table_attach (table, GTK_WIDGET (center),
			1, 2, 1, 2,
			GTK_EXPAND | GTK_FILL,
			GTK_EXPAND | GTK_FILL,
			0, 0);
	gtk_table_attach (table, GTK_WIDGET (right),
			2, 3, 1, 2,
			(GtkAttachOptions)0,
			GTK_FILL | GTK_EXPAND | GTK_SHRINK,
			0, 0);
	gtk_table_attach (table, GTK_WIDGET (bottom_left_corner),
			0, 1, 2, 3,
			(GtkAttachOptions)0, (GtkAttachOptions)0,
			0, 0);
	gtk_table_attach (table, GTK_WIDGET (bottom),
			1, 2, 2, 3,
			GTK_FILL | GTK_EXPAND | GTK_SHRINK,
			(GtkAttachOptions)0,
			0, 0);
	gtk_table_attach (table, GTK_WIDGET (bottom_right_corner),
			2, 3, 2, 3,
			(GtkAttachOptions)0, (GtkAttachOptions)0,
			0, 0);

	window->private->icon = NULL;
	window->private->icon_holder = gtk_event_box_new ();
	gtk_event_box_set_visible_window (
		GTK_EVENT_BOX (window->private->icon_holder), FALSE);
	//gtk_mdi_window_set_icon_from_stock (window, "gtk-file");
	window->private->title_bar = GTK_HBOX (gtk_hbox_new (FALSE, 0));
	gtk_widget_show (GTK_WIDGET (window->private->title_bar));

	window->private->title_label = GTK_LABEL (gtk_label_new (""));
	gtk_widget_show (GTK_WIDGET (window->private->title_label));

	/* titlebar */
	gtk_box_pack_start (GTK_BOX (window->private->title_bar),
			GTK_WIDGET (window->private->icon_holder),
			FALSE, FALSE, 0);
	gtk_box_pack_start (GTK_BOX (window->private->title_bar),
			GTK_WIDGET (window->private->title_label),
			TRUE, TRUE, 0);
	gtk_box_pack_start (GTK_BOX (window->private->title_bar),
			GTK_WIDGET (icon_butt),
			FALSE, FALSE, 0);
	gtk_box_pack_start (GTK_BOX (window->private->title_bar),
			GTK_WIDGET (max_butt),
			FALSE, FALSE, 0);
	gtk_box_pack_start (GTK_BOX (window->private->title_bar),
			GTK_WIDGET (close_butt),
			FALSE, FALSE, 0);
	/* -- titlebar --*/

	gtk_container_add (GTK_CONTAINER (window), GTK_WIDGET (table));

	gtk_box_pack_start (GTK_BOX (center),
			GTK_WIDGET (window->private->title_bar),
			FALSE, FALSE, 2);

	window->private->container =
			GTK_EVENT_BOX (gtk_event_box_new ());
	gtk_box_pack_start (GTK_BOX (center),
			GTK_WIDGET (window->private->container),
			TRUE, TRUE, 0);

	window->private->borders = g_slist_alloc ();

	window->private->borders->data = top_left_corner;
	g_slist_append (window->private->borders, top);
	g_slist_append (window->private->borders, top_right_corner);
	g_slist_append (window->private->borders, left);
	g_slist_append (window->private->borders, right);
	g_slist_append (window->private->borders, bottom_left_corner);
	g_slist_append (window->private->borders, bottom);
	g_slist_append (window->private->borders, bottom_right_corner);
}

static void gtk_mdi_window_finalize (GObject *object)
{
	if (G_OBJECT_CLASS (parent_class)->finalize)
		(* G_OBJECT_CLASS (parent_class)->finalize) (object);
}

GtkWidget *
gtk_mdi_window_new (void)
{
	GtkMdiWindow *new_mdi_window;
	GdkColor bg;

	new_mdi_window = (GtkMdiWindow *)gtk_widget_new (GTK_MDI_WINDOW_TYPE,
							NULL);
	gtk_event_box_set_visible_window (GTK_EVENT_BOX (new_mdi_window), TRUE);

	memset (&bg, 0, sizeof (bg));
	gdk_color_parse ("blue", &bg);
	gtk_widget_modify_bg (new_mdi_window, GTK_STATE_NORMAL, &bg);
	/*
	gtk_container_set_border_width (GTK_CONTAINER (new_mdi_window),
					4);
	*/

	return GTK_WIDGET (new_mdi_window);
}

void
gtk_mdi_window_maximize (GtkMdiWindow *window)
{
	guint width;
	guint height;
	GtkLayout *layout;
	GtkLayoutChild *child = (GtkLayoutChild *)window;

	g_return_if_fail (GTK_IS_MDI_WINDOW (window));

	layout = GTK_LAYOUT (window->mdicontainer);

	gdk_drawable_get_size ( GTK_WIDGET (window)->window,
				&(window->private->old_width),
				&(window->private->old_height));
	window->private->old_pos_x = child->x;
	window->private->old_pos_y = child->y;
	gtk_layout_get_size (layout, &width, &height);
	gtk_layout_move (layout, GTK_WIDGET (window), 0, 0);
	gtk_widget_set_size_request (GTK_WIDGET (window), width, height);

	window->private->old_state = window->private->state;
	window->private->state = GTK_MDI_WINDOW_MAXIMIZED;

	g_signal_emit (window->mdicontainer, gtk_mdi_signals[WINDOW_MAXIMIZED],
			0, window);
}

void
gtk_mdi_window_unmaximize (GtkMdiWindow *window)
{
	GtkLayout *layout;

	g_return_if_fail (GTK_IS_MDI_WINDOW (window));

	if (gtk_mdi_window_get_state (window) == GTK_MDI_WINDOW_MAXIMIZED)
	{
		gtk_layout_move (layout, GTK_WIDGET (window),
				window->private->old_pos_x,
				window->private->old_pos_y);

		gtk_widget_set_size_request (GTK_WIDGET (window),
				window->private->old_width,
				window->private->old_height);

		window->private->state = window->private->old_state;
		window->private->old_state = WINDOW_MAXIMIZED;

		g_signal_emit (window->mdicontainer,
				gtk_mdi_signals[WINDOW_UNMAXIMIZED], 0,
				window);
	}
}

void
gtk_mdi_window_iconify (GtkMdiWindow *window)
{
	g_return_if_fail (GTK_IS_MDI_WINDOW (window));

	window->private->old_state = window->private->state;
	window->private->state = GTK_MDI_WINDOW_ICONIFIED;
	gtk_widget_hide (GTK_WIDGET (window));
	g_signal_emit (window->mdicontainer, gtk_mdi_signals[WINDOW_ICONIFIED],
			0, window);
}

void
gtk_mdi_window_deiconify (GtkMdiWindow *window)
{
	g_return_if_fail (GTK_IS_MDI_WINDOW (window));

	if (gtk_mdi_window_get_state (window) == GTK_MDI_WINDOW_ICONIFIED)
	{
		gtk_widget_show (GTK_WIDGET (window));
		window->private->state = window->private->old_state;
		window->private->old_state = GTK_MDI_WINDOW_ICONIFIED;
		g_signal_emit (window->mdicontainer,
				gtk_mdi_signals[WINDOW_DEICONIFIED], 0, 
				window);
	}
}

void
gtk_mdi_window_close (GtkMdiWindow *window)
{
	GtkMdiWindowId id;

	g_return_if_fail (GTK_IS_MDI_WINDOW (window));

	id = window->id;
	g_object_unref (window);
	g_signal_emit (window->mdicontainer,
			gtk_mdi_signals[WINDOW_CLOSED], 0, id);
}

void
gtk_mdi_window_set_title (GtkMdiWindow *window, const gchar *title)
{
	g_return_if_fail (GTK_IS_MDI_WINDOW (window));
	g_return_if_fail (GTK_IS_MDI (window->mdicontainer));

	gtk_label_set_text (window->private->title_label, title);
	g_signal_emit (window->mdicontainer,
			gtk_mdi_signals[TITLE_CHANGED], 0, window);
}

void
gtk_mdi_window_set_resizeable (GtkMdiWindow *window, gboolean resizeable)
{
	gboolean old_value;

	g_return_if_fail (GTK_IS_MDI_WINDOW (window));
	g_return_if_fail (GTK_IS_MDI (window->mdicontainer));

	old_value = window->private->resizeable;

	window->private->resizeable = resizeable;
	if (!resizeable) _gtk_mdi_window_hide_borders (window);
	else _gtk_mdi_window_show_borders (window);
	g_signal_emit (window->mdicontainer,
		gtk_mdi_signals[WINDOW_PROPERTY_CHANGED], 0, window,
		"resizeable", old_value, resizeable);
}

gboolean
gtk_mdi_window_get_resizeable (GtkMdiWindow *window)
{
	return window->private->resizeable;
}

void
gtk_mdi_window_move (GtkMdiWindow *window, guint x, guint y)
{
	GtkLayoutChild *child = (GtkLayoutChild *) window;

	g_return_if_fail (GTK_IS_MDI_WINDOW (window));

	window->private->old_pos_x = child->x;
	window->private->old_pos_y = child->y;
	gtk_layout_move (GTK_LAYOUT (window->mdicontainer),
			GTK_WIDGET (window), x, y);
}

void
gtk_mdi_window_get_position (GtkMdiWindow *window, guint *x, guint *y)
{
	GtkLayoutChild *child = (GtkLayoutChild *) window;

	g_return_if_fail (GTK_IS_MDI_WINDOW (window));

	*x = child->x;
	*y = child->y;
}

void
gtk_mdi_window_grab_focus (GtkLayoutChild *child)
{
	GtkMdiWindowPrivate *priv;
	GtkMdiWindow *window;

	g_return_if_fail (child != NULL);
	window = GTK_MDI_WINDOW (child->widget);
	g_return_if_fail (GTK_IS_MDI_WINDOW (window));

	priv = window->private;

	gdk_window_raise (child->widget->window);
}

void
gtk_mdi_window_resize (GtkMdiWindow *window, gint width, gint height)
{
	g_return_if_fail (GTK_IS_MDI_WINDOW (window));
	gtk_widget_set_size_request (GTK_WIDGET (window), width, height);
	g_signal_emit (window->mdicontainer, gtk_mdi_signals[WINDOW_RESIZED],
			0, window);
}

void
gtk_mdi_window_set_decorated (GtkMdiWindow *window, gboolean decorated)
{
	gboolean old_value;

	g_return_if_fail (GTK_IS_MDI_WINDOW (window));
	g_return_if_fail (GTK_IS_MDI (window->mdicontainer));

	old_value = window->private->decorated;
	window->private->decorated = decorated;
	gtk_mdi_window_set_resizeable (window, decorated);
	if (!decorated) _gtk_mdi_window_hide_titlebar (window);
	else _gtk_mdi_window_show_titlebar (window);
	g_signal_emit (window->mdicontainer,
		gtk_mdi_signals[WINDOW_PROPERTY_CHANGED], 0, window,
		"decorated", old_value, decorated);
}

gboolean
gtk_mdi_window_get_decorated (GtkMdiWindow *window)
{
	return window->private->decorated;
}

gboolean
gtk_mdi_window_has_focus (GtkMdiWindow *window)
{
	return window->private->has_focus;
}

void
gtk_mdi_window_set_has_focus (GtkMdiWindow *window, gboolean setting)
{
	gboolean old_value;

	g_return_if_fail (GTK_IS_MDI_WINDOW (window));
	g_return_if_fail (GTK_IS_MDI (window->mdicontainer));

	old_value = window->private->has_focus;
	window->private->has_focus = setting;
	g_signal_emit (window->mdicontainer,
		gtk_mdi_signals[WINDOW_PROPERTY_CHANGED], 0, window,
		"has_focus", old_value, setting);
}
void
gtk_mdi_set_skip_taskbar_hint (GtkMdiWindow *window, gboolean setting)
{
	g_return_if_fail (GTK_IS_MDI_WINDOW (window));
	window->private->skip_taskbar = setting;
}

gboolean
gtk_mdi_get_skip_taskbar_hint (GtkMdiWindow *window)
{
	return window->private->skip_taskbar;
}

void
gtk_mdi_set_skip_pager_hint (GtkMdiWindow *window, gboolean setting)
{
	g_return_if_fail (GTK_IS_MDI_WINDOW (window));
	window->private->skip_pager = setting;
}

gboolean
gtk_mdi_get_skip_pager_hint (GtkMdiWindow *window)
{
	return window->private->skip_pager;
}

void
gtk_mdi_set_accept_focus_hint (GtkMdiWindow *window, gboolean setting)
{
	g_return_if_fail (GTK_IS_MDI_WINDOW (window));
	window->private->accept_focus = setting;
}

gboolean
gtk_mdi_get_accept_focus_hint (GtkMdiWindow *window)
{
	return window->private->accept_focus;
}

void
gtk_mdi_window_get_size (GtkMdiWindow *window, gint *width, gint *height)
{
	GdkWindow *win;

	g_return_if_fail (GTK_IS_MDI_WINDOW (window));

	win = gtk_widget_get_parent_window (GTK_WIDGET (window));
	gdk_drawable_get_size (win, width, height);
}

GtkWidget *
gtk_mdi_window_get_icon (GtkMdiWindow *window)
{
	return window->private->icon;
}

void
gtk_mdi_window_set_icon (GtkMdiWindow *window, GtkWidget *icon)
{
	g_return_if_fail (GTK_IS_MDI_WINDOW (window));
	g_return_if_fail (GTK_IS_IMAGE (icon));
	g_return_if_fail (GTK_IS_MDI (window->mdicontainer));

	if (GTK_IS_IMAGE (window->private->icon))
		gtk_widget_destroy (window->private->icon);
	window->private->icon = GTK_IMAGE (icon);
	gtk_container_add (GTK_CONTAINER (window->private->icon_holder), icon);
	g_signal_emit (window->mdicontainer, gtk_mdi_signals[ICON_CHANGED], 0,
			window);
}

void
gtk_mdi_window_set_icon_from_file (GtkMdiWindow *window, const gchar *filename)
{
	GtkWidget *icon;

	g_return_if_fail (GTK_IS_MDI_WINDOW (window));
	g_return_if_fail (filename != NULL);

	icon = gtk_image_new_from_file (filename);
	gtk_mdi_window_set_icon (window, icon);
}

void
gtk_mdi_window_set_icon_from_stock (GtkMdiWindow *window, const gchar *stock_id)
{
	GtkImage *image;

	g_return_if_fail (GTK_IS_MDI_WINDOW (window));
	g_return_if_fail (stock_id != NULL);

	image = GTK_IMAGE (gtk_image_new_from_stock (stock_id,
			GTK_ICON_SIZE_MENU));

	gtk_mdi_window_set_icon (window, GTK_WIDGET (image));
	//g_object_unref (G_OBJECT (image));
}

G_CONST_RETURN gchar *
gtk_mdi_window_get_title (GtkMdiWindow *window)
{
	return gtk_label_get_text (window->private->title_label);
}

GtkMdiWindowState
gtk_mdi_window_get_state (GtkMdiWindow *window)
{
	return window->private->state;
}

static void
_gtk_mdi_window_show_borders (GtkMdiWindow *window)
{
	g_slist_foreach (window->private->borders, (GFunc)gtk_widget_show,
			NULL);
}

static void
_gtk_mdi_window_hide_borders (GtkMdiWindow *window)
{
	g_slist_foreach (window->private->borders, (GFunc)gtk_widget_hide,
			NULL);
}

static void
_gtk_mdi_window_show_titlebar (GtkMdiWindow *window)
{
	gtk_widget_show (GTK_WIDGET (window));
}

static void
_gtk_mdi_window_hide_titlebar (GtkMdiWindow *window)
{
	gtk_widget_hide (GTK_WIDGET (window));
}

static void
_gtk_mdi_window_load_borders (GtkEventBox *tlc, GtkEventBox *top,
			GtkEventBox *trc, GtkEventBox *left,
			GtkEventBox *right, GtkEventBox *blc,
			GtkEventBox *bottom, GtkEventBox *brc)
{
	GtkWidget *image;

	image = gtk_image_new_from_file ("tlc.png");
	gtk_container_add (GTK_CONTAINER (tlc), image);

	image = gtk_image_new_from_file ("top.png");
	gtk_container_add (GTK_CONTAINER (top), image);

	image = gtk_image_new_from_file ("trc.png");
	gtk_container_add (GTK_CONTAINER (trc), image);

	image = gtk_image_new_from_file ("left.png");
	gtk_container_add (GTK_CONTAINER (left), image);

	image = gtk_image_new_from_file ("right.png");
	gtk_container_add (GTK_CONTAINER (right), image);

	image = gtk_image_new_from_file ("blc.png");
	gtk_container_add (GTK_CONTAINER (blc), image);

	image = gtk_image_new_from_file ("bottom.png");
	gtk_container_add (GTK_CONTAINER (bottom), image);

	image = gtk_image_new_from_file ("tlc.png");
	gtk_container_add (GTK_CONTAINER (brc), image);
}

gboolean
_gtk_mdi_window_maximize (GtkWidget *widget, GtkWidget *data)
{
	gtk_mdi_window_maximize (GTK_MDI_WINDOW (data));
	return FALSE;
}

gboolean
_gtk_mdi_window_close (GtkWidget *widget, GtkWidget *data)
{
	gtk_mdi_window_close (GTK_MDI_WINDOW (data));
	return FALSE;
}

gboolean
_gtk_mdi_window_iconify (GtkWidget *widget, GtkWidget *data)
{
	gtk_mdi_window_iconify (GTK_MDI_WINDOW (data));
	return FALSE;
}

